# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow solutions in laplace space.

.. currentmodule:: anaflow.flow.laplace

The following functions are provided

.. autosummary::
   lap_trans_flow_cyl
   grf_laplace
"""

from __future__ import absolute_import, division, print_function

import warnings

import numpy as np
import scipy.sparse as sps
from scipy.sparse.linalg import spsolve
from scipy.special import k0, k1, kn, kv, i0, i1, iv, gamma
from anaflow.tools.special import sph_surf

# ignore Umpfack warnings for almost singular matrices
try:
    # first look if the umfpack is available
    from scikits.umfpack import UmfpackWarning

    SlvWarn = UmfpackWarning
except ImportError:
    # if umfpack is not present, use scipy.sparse MatrixRankWarning
    SlvWarn = sps.linalg.MatrixRankWarning

__all__ = ["lap_trans_flow_cyl", "grf_laplace"]


def get_bessel(nu):
    """Get the right bessel functions for the GRF-model"""
    if np.isclose(nu, 0):
        kv0 = lambda x: k0(x)
        kv1 = lambda x: k1(x)
        iv0 = lambda x: i0(x)
        iv1 = lambda x: i1(x)
    if np.isclose(nu, np.around(nu)):
        kv0 = lambda x: kn(int(nu), x)
        kv1 = lambda x: kn(int(nu - 1), x)
        iv0 = lambda x: iv(int(nu), x)
        iv1 = lambda x: iv(int(nu - 1), x)
    else:
        kv0 = lambda x: kv(nu, x)
        kv1 = lambda x: kv(nu - 1, x)
        iv0 = lambda x: iv(nu, x)
        iv1 = lambda x: iv(nu - 1, x)
    return kv0, kv1, iv0, iv1


###############################################################################
# The generic solver of the 2D radial transient groundwaterflow equation
# in Laplace-space with a pumping condition and a fix zero boundary-head
###############################################################################


def lap_trans_flow_cyl(
    s, rad=None, rpart=None, Spart=None, Tpart=None, Qw=None, Twell=None
):
    """
    A diskmodel for transient flow in Laplace-space

    The solution of the diskmodel for transient flow under a pumping condition
    in a confined aquifer in Laplace-space.
    The solutions assumes concentric disks around the pumpingwell,
    where each disk has its own transmissivity and storativity value.

    Parameters
    ----------
    s : :class:`numpy.ndarray`
        Array with all Laplace-space-points
        where the function should be evaluated
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    rpart : :class:`numpy.ndarray`
        Given radii separating the disks as well as starting- and endpoints
    Tpart : :class:`numpy.ndarray`
        Given transmissivity values for each disk
    Spart : :class:`numpy.ndarray`
        Given storativity values for each disk
    Qw : :class:`float`
        Pumpingrate at the well
    Twell : :class:`float`, optional
        Transmissivity at the well. Default: ``Tpart[0]``

    Returns
    -------
    lap_transgwflow_cyl : :class:`numpy.ndarray`
        Array with all values in laplace-space

    Examples
    --------
    >>> lap_transgwflow_cyl([5,10],[1,2,3],[0,2,10],[1e-3,1e-3],[1e-3,2e-3],-1)
    array([[ -2.71359196e+00,  -1.66671965e-01,  -2.82986917e-02],
           [ -4.58447458e-01,  -1.12056319e-02,  -9.85673855e-04]])
    """

    # ensure that input is treated as arrays
    s = np.squeeze(s).reshape(-1)
    rad = np.squeeze(rad).reshape(-1)
    rpart = np.squeeze(rpart).reshape(-1)
    Spart = np.squeeze(Spart).reshape(-1)
    Tpart = np.squeeze(Tpart).reshape(-1)

    # get the number of partitions
    parts = len(Tpart)

    # initialize the result
    res = np.zeros(s.shape + rad.shape)

    # set the general pumping-condtion
    if Twell is None:
        Twell = Tpart[0]
    Q = Qw / (2.0 * np.pi * Twell)

    # if there is a homgeneouse aquifer, compute the result by hand
    if parts == 1:
        # calculate the square-root of the diffusivities
        difsr = np.sqrt(Spart[0] / Tpart[0])

        for si, se in np.ndenumerate(s):
            Cs = np.sqrt(se) * difsr

            # set the pumping-condition at the well
            Qs = Q / se

            # incorporate the boundary-conditions
            if rpart[0] == 0.0:
                Bs = Qs
                if rpart[-1] == np.inf:
                    As = 0.0
                else:
                    As = -Qs * k0(Cs * rpart[-1]) / i0(Cs * rpart[-1])

            else:
                if rpart[-1] == np.inf:
                    As = 0.0
                    Bs = Qs / (Cs * rpart[0] * k1(Cs * rpart[0]))
                else:
                    det = i1(Cs * rpart[0]) * k0(Cs * rpart[-1]) + k1(
                        Cs * rpart[0]
                    ) * i0(Cs * rpart[-1])
                    As = -Qs / (Cs * rpart[0]) * k0(Cs * rpart[-1]) / det
                    Bs = Qs / (Cs * rpart[0]) * i0(Cs * rpart[-1]) / det

            # calculate the head
            for ri, re in np.ndenumerate(rad):
                if re < rpart[-1]:
                    res[si + ri] = As * i0(Cs * re) + Bs * k0(Cs * re)

    # if there is more than one partition, create an equation system
    else:
        # initialize LHS and RHS for the linear equation system
        # Mb is the banded matrix for the Eq-System
        V = np.zeros(2 * (parts))
        Mb = np.zeros((5, 2 * (parts)))
        # the positions of the diagonals of the matrix set in Mb
        diagpos = [2, 1, 0, -1, -2]
        # set the standard boundary conditions for rwell=0.0 and rinf=np.inf
        Mb[2, 0] = 1.0
        Mb[-3, -1] = 1.0

        # calculate the consecutive fractions of the transmissivities
        Tfrac = Tpart[:-1] / Tpart[1:]

        # calculate the square-root of the diffusivities
        difsr = np.sqrt(Spart / Tpart)

        # calculate a temporal substitution
        tmp = Tfrac * difsr[:-1] / difsr[1:]

        # match the radii to the different disks
        pos = np.searchsorted(rpart, rad) - 1

        # iterate over the laplace-variable
        for si, se in enumerate(s):
            Cs = np.sqrt(se) * difsr

            # set the pumping-condition at the well
            # --> implement other pumping conditions
            V[0] = Q / se

            # set the boundary-conditions if needed
            if rpart[0] > 0.0:
                Mb[2, 0] = Cs[0] * rpart[0] * k1(Cs[0] * rpart[0])
                Mb[1, 1] = -Cs[0] * rpart[0] * i1(Cs[0] * rpart[0])
            if rpart[-1] < np.inf:
                Mb[-2, -2] = k0(Cs[-1] * rpart[-1])
                Mb[-3, -1] = i0(Cs[-1] * rpart[-1])

            # generate the equation system as banded matrix
            for i in range(parts - 1):
                Mb[0, 2 * i + 3] = -i0(Cs[i + 1] * rpart[i + 1])
                Mb[1, 2 * i + 2 : 2 * i + 4] = [
                    -k0(Cs[i + 1] * rpart[i + 1]),
                    -i1(Cs[i + 1] * rpart[i + 1]),
                ]
                Mb[2, 2 * i + 1 : 2 * i + 3] = [
                    i0(Cs[i] * rpart[i + 1]),
                    k1(Cs[i + 1] * rpart[i + 1]),
                ]
                Mb[3, 2 * i : 2 * i + 2] = [
                    k0(Cs[i] * rpart[i + 1]),
                    tmp[i] * i1(Cs[i] * rpart[i + 1]),
                ]
                Mb[4, 2 * i] = -tmp[i] * k1(Cs[i] * rpart[i + 1])

            # genearate the cooeficient matrix as a spare matrix
            M = sps.spdiags(Mb, diagpos, 2 * parts, 2 * parts, format="csc")

            # solve the Eq-Sys and ignore errors from the umf-pack
            with warnings.catch_warnings():
                # warnings.simplefilter("ignore")
                warnings.simplefilter("ignore", SlvWarn)
                warnings.simplefilter("ignore", RuntimeWarning)
                X = spsolve(M, V, use_umfpack=True)

            # to suppress numerical errors, set NAN values to 0
            X[np.logical_not(np.isfinite(X))] = 0.0

            # calculate the head
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                res[si, :] = X[2 * pos] * k0(Cs[pos] * rad) + X[
                    2 * pos + 1
                ] * i0(Cs[pos] * rad)

        # set problematic values to 0
        # --> the algorithm tends to violate small values,
        #     therefore this approachu is suitable
        res[np.logical_not(np.isfinite(res))] = 0.0

    return res


###############################################################################
# The generic solver of the 2D radial transient groundwaterflow equation
# in Laplace-space with a pumping condition and a fix zero boundary-head
###############################################################################


def grf_laplace(
    s,
    rad=None,
    dim=2,
    lat_ext=1.0,
    rpart=None,
    Spart=None,
    Kpart=None,
    Qw=None,
    Kwell=None,
):
    """
    A modified GRF-model for transient flow in Laplace-space

    The General Radial Flow (GRF) Model allowes fractured dimensions for
    transient flow under a pumping condition
    in a confined aquifer in Laplace-space.
    The solutions assumes concentric disks around the pumpingwell,
    where each disk has its own conductivity and storativity value.

    Parameters
    ----------
    s : :class:`numpy.ndarray`
        Array with all Laplace-space-points
        where the function should be evaluated
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    dim : :class:`float`
        Flow dimension. Default: 3
    lat_ext : :class:`float`
        The lateral extend of the flow-domain in `L^(3-dim)`. Default: 1
    rpart : :class:`numpy.ndarray` of length N+1
        Given radii separating the disks as well as starting- and endpoints
    Kpart : :class:`numpy.ndarray` of length N
        Given conductivity values for each disk
    Spart : :class:`numpy.ndarray` of length N
        Given storativity values for each disk
    Qw : :class:`float`
        Pumpingrate at the well
    Twell : :class:`float`, optional
        Transmissivity at the well. Default: ``Tpart[0]``

    Returns
    -------
    grf_laplace : :class:`numpy.ndarray`
        Array with all values in laplace-space

    Examples
    --------
    >>> grf_laplace([5,10],[1,2,3], 2, 1, [0,2,10],[1e-3,1e-3],[1e-3,2e-3],-1)
    array([[ -2.71359196e+00,  -1.66671965e-01,  -2.82986917e-02],
           [ -4.58447458e-01,  -1.12056319e-02,  -9.85673855e-04]])
    """

    # ensure that input is treated as arrays
    s = np.squeeze(s).reshape(-1)
    rad = np.squeeze(rad).reshape(-1)
    rpart = np.squeeze(rpart).reshape(-1)
    Spart = np.squeeze(Spart).reshape(-1)
    Kpart = np.squeeze(Kpart).reshape(-1)

    # the dimension is used by nu in the literature (See Barker 88)
    dim = float(dim)
    nu = 1.0 - dim / 2.0
    # the lateral extend is a bit subtle in fractured dimension
    lat_ext = float(lat_ext)
    Qw = float(Qw)

    # get the number of partitions
    parts = len(Kpart)

    # initialize the result
    res = np.zeros(s.shape + rad.shape)

    # set the conductivity at the well
    if Kwell is None:
        Kwell = Kpart[0]

    # the first sqrt of the diffusivity values
    diff_sr0 = np.sqrt(Spart[0] / Kpart[0])
    # the pumping-condition
    Q = Qw / (Kwell * sph_surf(dim) * lat_ext ** (3.0 - dim))

    # set the general pumping-condtion depending on the well-radius
    if rpart[0] > 0.0:
        Q /= -diff_sr0 * rpart[0] ** (1 - nu)
        Qs = Q * s ** (-1.5)
    else:
        Q *= (2 / diff_sr0) ** nu / gamma(1 - nu)
        Qs = Q * s ** (-nu / 2 - 1)

    # get the right modified bessel-functions according to the dimension
    # Jv0 = J(v) ; Jv1 = J(v-1) for J in [k, i]
    kv0, kv1, iv0, iv1 = get_bessel(nu)

    # if there is a homgeneouse aquifer, compute the result by hand
    if parts == 1:
        # calculate the square-root of the diffusivities
        difsr = np.sqrt(Spart[0] / Kpart[0])

        # initialize the equation system
        V = np.zeros(2, dtype=float)
        M = np.eye(2, dtype=float)

        for si, se in enumerate(s):
            Cs = np.sqrt(se) * difsr

            # set the pumping-condition at the well
            V[0] = Qs[si]

            # incorporate the boundary-conditions
            if rpart[0] > 0.0:
                M[0, :] = [-kv1(Cs * rpart[0]), iv1(Cs * rpart[0])]
            if rpart[-1] < np.inf:
                M[1, :] = [kv0(Cs * rpart[-1]), iv0(Cs * rpart[-1])]

            # solve the equation system
            As, Bs = np.linalg.solve(M, V)

            # calculate the head
            for ri, re in enumerate(rad):
                if re < rpart[-1]:
                    res[si, ri] = re ** nu * (
                        As * kv0(Cs * re) + Bs * iv0(Cs * re)
                    )

    # if there is more than one partition, create an equation system
    else:
        # initialize LHS and RHS for the linear equation system
        # Mb is the banded matrix for the Eq-System
        V = np.zeros(2 * (parts))
        Mb = np.zeros((5, 2 * (parts)))
        # the positions of the diagonals of the matrix set in Mb
        diagpos = [2, 1, 0, -1, -2]
        # set the standard boundary conditions for rwell=0.0 and rinf=np.inf
        Mb[2, 0] = 1.0
        Mb[-3, -1] = 1.0

        # calculate the consecutive fractions of the conductivities
        Kfrac = Kpart[:-1] / Kpart[1:]

        # calculate the square-root of the diffusivities
        difsr = np.sqrt(Spart / Kpart)

        # calculate a temporal substitution (factor from mass-conservation)
        tmp = Kfrac * difsr[:-1] / difsr[1:]

        # match the radii to the different disks
        pos = np.searchsorted(rpart, rad) - 1

        # iterate over the laplace-variable
        for si, se in enumerate(s):
            Cs = np.sqrt(se) * difsr

            # set the pumping-condition at the well
            # --> implement other pumping conditions
            V[0] = Qs[si]

            # set the boundary-conditions if needed
            if rpart[0] > 0.0:
                Mb[2, 0] = -kv1(Cs[0] * rpart[0])
                Mb[1, 1] = iv1(Cs[0] * rpart[0])
            if rpart[-1] < np.inf:
                Mb[-2, -2] = kv0(Cs[-1] * rpart[-1])
                Mb[-3, -1] = iv0(Cs[-1] * rpart[-1])

            # generate the equation system as banded matrix
            for i in range(parts - 1):
                Mb[0, 2 * i + 3] = -iv0(Cs[i + 1] * rpart[i + 1])
                Mb[1, 2 * i + 2 : 2 * i + 4] = [
                    -kv0(Cs[i + 1] * rpart[i + 1]),
                    -iv1(Cs[i + 1] * rpart[i + 1]),
                ]
                Mb[2, 2 * i + 1 : 2 * i + 3] = [
                    iv0(Cs[i] * rpart[i + 1]),
                    kv1(Cs[i + 1] * rpart[i + 1]),
                ]
                Mb[3, 2 * i : 2 * i + 2] = [
                    kv0(Cs[i] * rpart[i + 1]),
                    tmp[i] * iv1(Cs[i] * rpart[i + 1]),
                ]
                Mb[4, 2 * i] = -tmp[i] * kv1(Cs[i] * rpart[i + 1])

            # genearate the cooeficient matrix as a spare matrix
            M = sps.spdiags(Mb, diagpos, 2 * parts, 2 * parts, format="csc")

            # solve the Eq-Sys and ignore errors from the umf-pack
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", SlvWarn)
                warnings.simplefilter("ignore", RuntimeWarning)
                X = spsolve(M, V, use_umfpack=True)

            # to suppress numerical errors, set NAN values to 0
            # --> the algorithm tends to violate small values,
            #     therefore this approachu is suitable
            X[np.logical_not(np.isfinite(X))] = 0.0

            # calculate the head
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                res[si, :] = rad ** nu * (
                    X[2 * pos] * kv0(Cs[pos] * rad)
                    + X[2 * pos + 1] * iv0(Cs[pos] * rad)
                )

        # set problematic values to 0
        # --> the algorithm tends to violate small values,
        #     therefore this approachu is suitable
        res[np.logical_not(np.isfinite(res))] = 0.0

    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow solutions in homogeneous aquifers.

.. currentmodule:: anaflow.flow.homogeneous

The following functions are provided

.. autosummary::
   thiem
   theis
"""

from __future__ import absolute_import, division, print_function

import numpy as np

from anaflow.tools.laplace import stehfest as sf
from anaflow.tools.special import well_solution, inc_gamma
from anaflow.flow.laplace import lap_trans_flow_cyl, grf_laplace

__all__ = ["thiem", "theis"]


###############################################################################
# Thiem-solution
###############################################################################


def thiem(rad, Rref, T, Qw, href=0.0):
    """
    The Thiem solution

    The Thiem solution for steady-state flow under a pumping condition
    in a confined and homogeneous aquifer.
    This solution was presented in ''Thiem 1906''[R6]_.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    Rref : :class:`float`
        Reference radius with known head (see `href`)
    T : :class:`float`
        Given transmissivity of the aquifer
    Qw : :class:`float`
        Pumpingrate at the well
    href : :class:`float`, optional
        Reference head at the reference-radius `Rref`. Default: ``0.0``

    Returns
    -------
    thiem : :class:`numpy.ndarray`
        Array with all heads at the given radii.

    References
    ----------
    .. [R6] Thiem, G.,
       ''Hydrologische Methoden, J.M. Gebhardt'', Leipzig, 1906.

    Notes
    -----
    The parameters ``rad``, ``Rref`` and ``T`` will be checked for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> thiem([1,2,3], 10, 0.001, -0.001)
    array([-0.3664678 , -0.25615   , -0.19161822])
    """

    rad = np.squeeze(rad)

    # check the input
    if Rref <= 0.0:
        raise ValueError("The reference-radius needs to be greater than 0")
    if np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if T <= 0.0:
        raise ValueError("The Transmissivity needs to be positiv")

    return -Qw / (2.0 * np.pi * T) * np.log(rad / Rref) + href


###############################################################################
# Theis-solution
###############################################################################


def theis(
    rad,
    time,
    T,
    S,
    Qw,
    struc_grid=True,
    rwell=0.0,
    rinf=np.inf,
    hinf=0.0,
    stehfestn=12,
):
    """
    The Theis solution

    The Theis solution for transient flow under a pumping condition
    in a confined and homogeneous aquifer.
    This solution was presented in ''Theis 1935''[Theis35]_.

    .. [Theis35] Theis, C.,
       ''The relation between the lowering of the piezometric surface and the
       rate and duration of discharge of a well using groundwater storage'',
       Trans. Am. Geophys. Union, 16, 519-524, 1935

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    T : :class:`float`
        Given transmissivity of the aquifer
    S : :class:`float`
        Given storativity of the aquifer
    Qw : :class:`float`
        Pumpingrate at the well
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    rwell : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    rinf : :class:`float`, optional
        Radius of the outer boundary of the aquifer. Default: ``np.inf``
    hinf : :class:`float`, optional
        Reference head at the outer boundary `rinf`. Default: ``0.0``
    stehfestn : :class:`int`, optional
        If `rwell` or `rinf` are not default, the solution is calculated in
        Laplace-space. The back-transformation is performed with the stehfest-
        algorithm. Here you can specify the number of interations within this
        algorithm. Default: ``12``

    Returns
    -------
    theis : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Notes
    -----
    The parameters ``rad``, ``Rref`` and ``T`` will be checked for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> theis([1,2,3], [10,100], 0.001, 0.001, -0.001)
    array([[-0.24959541, -0.14506368, -0.08971485],
           [-0.43105106, -0.32132823, -0.25778313]])
    """

    # ensure that 'rad' and 'time' are arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)

    if not struc_grid:
        grid_shape = rad.shape
        rad = rad.reshape(-1)

    # check the input
    if rwell < 0.0:
        raise ValueError("The wellradius needs to be >= 0")
    if rinf <= rwell:
        raise ValueError(
            "The upper boundary needs to be greater than the wellradius"
        )
    if np.any(rad < rwell) or np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if np.any(time <= 0.0):
        raise ValueError("The given times need to be > 0")
    if not struc_grid and not rad.shape == time.shape:
        raise ValueError(
            "For unstructured grid the number of time- & radii-pts must equal"
        )
    if T <= 0.0:
        raise ValueError("The Transmissivity needs to be positiv")
    if S <= 0.0:
        raise ValueError("The Storage needs to be positiv")
    if not isinstance(stehfestn, int):
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be an integer"
        )
    if stehfestn <= 1:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be > 1"
        )
    if stehfestn % 2 != 0:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be even"
        )

    if rwell == 0.0 and rinf == np.inf:
        res = well_solution(rad, time, T, S, Qw)

    else:
        rpart = np.array([rwell, rinf])
        Tpart = np.array([T])
        Spart = np.array([S])

        # write the paramters in kwargs to use the stehfest-algorithm
        kwargs = {
            "rad": rad,
            "Qw": Qw,
            "rpart": rpart,
            "Spart": Spart,
            "Tpart": Tpart,
        }

        # call the stehfest-algorithm
        res = sf(lap_trans_flow_cyl, time, bound=stehfestn, **kwargs)

    # if the input are unstructured space-time points, return an array
    if not struc_grid and grid_shape:
        res = np.copy(np.diag(res).reshape(grid_shape))

    # add the reference head
    res += hinf

    return res


def grf_model(rad, time, K, S, Qw, dim=3, lat_ext=1.0):
    # ensure that 'rad' and 'time' are arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)

    res = np.zeros(time.shape + rad.shape)

    nu = 1.0 - dim / 2.0

    for ti, te in np.ndenumerate(time):
        for ri, re in np.ndenumerate(rad):
            u = S * re ** 2 / (4 * K * te)
            res[ti + ri] = (
                Qw / (4.0 * np.pi ** (1 - nu) * K * lat_ext) * re ** (2 * nu)
            )
            res[ti + ri] *= inc_gamma(-nu, u)
    return res


def grf_lap(rad, time, K, S, Qw, rpart=None, dim=3, lat_ext=1.0):
    # ensure that 'rad' and 'time' are arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)
    if rpart is None:
        rpart = np.array([0, np.inf])
    Kpart = np.array(K, ndmin=1)
    Spart = np.array(S, ndmin=1)
    kwargs = {
        "rad": rad,
        "Qw": Qw,
        "rpart": rpart,
        "Spart": Spart,
        "Kpart": Kpart,
        "dim": dim,
        "lat_ext": lat_ext,
    }

    # call the stehfest-algorithm
    res = sf(grf_laplace, time, bound=12, **kwargs)
    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()

    print(theis([1, 2, 3], [10, 100], 1e-4, 1e-4, 1e-4))
    print(grf_model([1, 2, 3], [10, 100], 1e-4, 1e-4, 1e-4, dim=1.2))
    print(grf_lap([1, 2, 3], [10, 100], 1e-4, 1e-4, 1e-4, dim=1.2))



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing special flow solutions.

.. currentmodule:: anaflow.flow.special

The following functions are provided

.. autosummary::
   diskmodel
   grf_model
"""

from __future__ import absolute_import, division, print_function

import numpy as np

from anaflow.tools.laplace import stehfest as sf
from anaflow.flow.laplace import lap_trans_flow_cyl, grf_laplace

__all__ = ["diskmodel", "grf_model"]


###############################################################################
# solution for a disk-model
###############################################################################


def diskmodel(
    rad,
    time,
    Tpart,
    Spart,
    Rpart,
    Qw,
    struc_grid=True,
    rwell=0.0,
    rinf=np.inf,
    hinf=0.0,
    stehfestn=12,
):
    """
    A diskmodel for transient flow

    A diskmodel for transient flow under a pumping condition
    in a confined aquifer. The solutions assumes concentric disks around the
    pumpingwell, where each disk has its own transmissivity and storativity
    value.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    Tpart : :class:`numpy.ndarray`
        Given transmissivity values for each disk
    Spart : :class:`numpy.ndarray`
        Given storativity values for each disk
    Rpart : :class:`numpy.ndarray`
        Given radii separating the disks
    Qw : :class:`float`
        Pumpingrate at the well
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    rwell : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    rinf : :class:`float`, optional
        Radius of the outer boundary of the aquifer. Default: ``np.inf``
    hinf : :class:`float`, optional
        Reference head at the outer boundary `rinf`. Default: ``0.0``
    stehfestn : :class:`int`, optional
        Since the solution is calculated in Laplace-space, the
        back-transformation is performed with the stehfest-algorithm.
        Here you can specify the number of interations within this
        algorithm. Default: ``12``

    Returns
    -------
    diskmodel : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Notes
    -----
    The parameters ``rad``, ``time``, ``Tpart`` and ``Spart`` will be checked
    for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> diskmodel([1,2,3], [10, 100], [1e-3, 2e-3], [1e-3, 1e-3], [2], -1e-3)
    array([[-0.20312814, -0.09605675, -0.06636862],
           [-0.29785979, -0.18784251, -0.15582597]])
    """

    # ensure that input is treated as arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)
    Tpart = np.array(Tpart)
    Spart = np.array(Spart)
    Rpart = np.array(Rpart)

    if not struc_grid:
        grid_shape = rad.shape
        rad = rad.reshape(-1)

    # check the input
    if rwell < 0.0:
        raise ValueError("The wellradius needs to be >= 0")
    if rinf <= rwell:
        raise ValueError(
            "The upper boundary needs to be greater than the wellradius"
        )
    if not all(Rpart[i] < Rpart[i + 1] for i in range(len(Rpart) - 1)):
        raise ValueError("The radii of the zones need to be sorted")
    if np.any(Rpart <= rwell):
        raise ValueError(
            "The radii of the zones need to be greater than the wellradius"
        )
    if np.any(Rpart >= rinf):
        raise ValueError(
            "The radii of the zones need to be less than the outer radius"
        )
    if np.any(rad < rwell) or np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if np.any(time <= 0.0):
        raise ValueError("The given times need to be >= 0")
    if not struc_grid and not rad.shape == time.shape:
        raise ValueError(
            "For unstructured grid the number of time- & radii-pts must equal"
        )
    if np.any(Tpart <= 0.0):
        raise ValueError("The Transmissivities need to be positiv")
    if np.any(Spart <= 0.0):
        raise ValueError("The Storages need to be positiv")
    if not isinstance(stehfestn, int):
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be an integer"
        )
    if stehfestn <= 1:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be > 1"
        )
    if stehfestn % 2 != 0:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be even"
        )

    rpart = np.append(np.array([rwell]), Rpart)
    rpart = np.append(rpart, np.array([rinf]))

    # write the paramters in kwargs to use the stehfest-algorithm
    kwargs = {
        "rad": rad,
        "Qw": Qw,
        "rpart": rpart,
        "Spart": Spart,
        "Tpart": Tpart,
    }

    # call the stehfest-algorithm
    res = sf(lap_trans_flow_cyl, time, bound=stehfestn, **kwargs)

    # if the input are unstructured space-time points, return an array
    if not struc_grid and grid_shape:
        res = np.copy(np.diag(res).reshape(grid_shape))

    # add the reference head
    res += hinf

    return res


def grf_model(
    rad,
    time,
    K_part,
    S_part,
    R_part,
    Qw,
    dim=2,
    lat_ext=1.0,
    struc_grid=True,
    r_well=0.0,
    r_bound=np.inf,
    head_bound=0.0,
    stehfest_n=12,
):
    """
    The extended "General radial flow" model for transient flow

    The general radial flow (GRF) model by Barker introduces an arbitrary
    dimension for radial groundwater flow. We introduced the possibility to
    define radial dependet conductivity and storage values.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    K_part : :class:`numpy.ndarray`
        Given conductivity values for each disk
    S_part : :class:`numpy.ndarray`
        Given storativity values for each disk
    R_part : :class:`numpy.ndarray`
        Given radii separating the disks
    Qw : :class:`float`
        Pumpingrate at the well
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    r_well : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    r_bound : :class:`float`, optional
        Radius of the outer boundary of the aquifer. Default: ``np.inf``
    head_bound : :class:`float`, optional
        Reference head at the outer boundary `rinf`. Default: ``0.0``
    stehfest_n : :class:`int`, optional
        Since the solution is calculated in Laplace-space, the
        back-transformation is performed with the stehfest-algorithm.
        Here you can specify the number of interations within this
        algorithm. Default: ``12``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Notes
    -----
    The parameters ``rad``, ``time``, ``K_part`` and ``S_part`` will be checked
    for positivity.
    """

    # ensure that input is treated as arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)
    K_part = np.array(K_part)
    S_part = np.array(S_part)
    R_part = np.array(R_part)

    if not struc_grid:
        grid_shape = rad.shape
        rad = rad.reshape(-1)

    rpart = np.append(np.array([r_well]), R_part)
    rpart = np.append(rpart, np.array([r_bound]))

    # write the paramters in kwargs to use the stehfest-algorithm
    kwargs = {
        "rad": rad,
        "Qw": Qw,
        "rpart": rpart,
        "Spart": S_part,
        "Kpart": K_part,
        "dim": dim,
        "lat_ext": lat_ext,
    }

    # call the stehfest-algorithm
    res = sf(grf_laplace, time, bound=stehfest_n, **kwargs)

    # if the input are unstructured space-time points, return an array
    if not struc_grid and grid_shape:
        res = np.copy(np.diag(res).reshape(grid_shape))

    # add the reference head
    res += head_bound

    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow-solutions for the groundwater flow equation.

Subpackages
^^^^^^^^^^^

.. currentmodule:: anaflow.flow

.. autosummary::
    homogeneous
    heterogeneous
    special
    laplace

Solutions
^^^^^^^^^

Homogeneous
~~~~~~~~~~~

.. currentmodule:: anaflow.flow.homogeneous

Solutions for homogeneous aquifers

.. autosummary::
   thiem
   theis

Heterogeneous
~~~~~~~~~~~~~

.. currentmodule:: anaflow.flow.heterogeneous

Solutions for heterogeneous aquifers

.. autosummary::
   ext_thiem2D
   ext_theis2D
   ext_thiem3D
   ext_theis3D

Special
~~~~~~~

.. currentmodule:: anaflow.flow.special

Special solutions for special aquifers

.. autosummary::
   grf_model
"""
from __future__ import absolute_import

from anaflow.flow.homogeneous import thiem, theis
from anaflow.flow.heterogeneous import (
    ext_thiem2D,
    ext_theis2D,
    ext_thiem3D,
    ext_theis3D,
)
from anaflow.flow.special import grf_model

__all__ = [
    "thiem",
    "theis",
    "ext_thiem2D",
    "ext_theis2D",
    "ext_thiem3D",
    "ext_theis3D",
    "grf_model",
]



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing miscellaneous tools.

Subpackages
^^^^^^^^^^^

.. currentmodule:: anaflow.tools

.. autosummary::
    laplace
    mean
    special
    coarse_graining
"""
from __future__ import absolute_import



# -*- coding: utf-8 -*-
"""
Purpose
=======

Anaflow provides several analytical and semi-analytical solutions for the
groundwater-flow-equation.

Subpackages
===========

.. autosummary::
    flow
    tools

Solutions
=========

Homogeneous
^^^^^^^^^^^

.. currentmodule:: anaflow.flow.homogeneous

Solutions for homogeneous aquifers

.. autosummary::
   thiem
   theis

Heterogeneous
^^^^^^^^^^^^^

.. currentmodule:: anaflow.flow.heterogeneous

Solutions for heterogeneous aquifers

.. autosummary::
   ext_thiem2D
   ext_theis2D
   ext_thiem3D
   ext_theis3D

Special
^^^^^^^

.. currentmodule:: anaflow.flow.special

Special solutions for special aquifers

.. autosummary::
   grf_model

Laplace
=======

.. currentmodule:: anaflow.tools.laplace

Helping functions related to the laplace-transformation

.. autosummary::
   get_lap
   get_lap_inv
"""
from __future__ import absolute_import

from anaflow._version import __version__

from anaflow.flow import (
    thiem,
    theis,
    ext_thiem2D,
    ext_theis2D,
    ext_thiem3D,
    ext_theis3D,
    grf_model,
)
from anaflow.tools.laplace import get_lap_inv, get_lap

__all__ = ["__version__"]

__all__ += [
    "thiem",
    "theis",
    "ext_thiem2D",
    "ext_theis2D",
    "ext_thiem3D",
    "ext_theis3D",
    "grf_model",
    "get_lap_inv",
    "get_lap",
]



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing special functions

.. currentmodule:: anaflow.tools.special

The following functions are provided

.. autosummary::

   sph_surf
   radii
   specialrange
   specialrange_cut
   aniso
   well_solution
   inc_gamma
"""

from __future__ import absolute_import, division, print_function

import numpy as np
from scipy.special import gamma, gammaincc, exp1, expn

__all__ = [
    "sph_surf",
    "radii",
    "specialrange",
    "specialrange_cut",
    "aniso",
    "well_solution",
    "inc_gamma",
]


def sph_surf(dim):
    """surface of the d-dimensional sphere"""
    return 2.0 * np.sqrt(np.pi) ** dim / gamma(dim / 2.0)


def radii(parts, rwell=0.0, rinf=np.inf, rlast=500.0, typ="log"):
    """
    Calculation of specific point distributions

    Calculation of specific point distributions for the diskmodel.

    Parameters
    ----------
    parts : :class:`int`
        Number of partitions.
    rwell : :class:`float`, optional
        Radius at the well. Default: ``0.0``
    rinf : :class:`float`, optional
        Radius at the outer boundary. Default: :any:``np.inf``
    rlast : :class:`float`, optional
        Setting the last radius befor the outer boundary. Default: ``500.0``
    typ : :class:`str`, optional
        Setting the distribution type of the radii. Default: ``"log"``

    Returns
    -------
    p_rad : :class:`numpy.ndarray`
        Array containing the separating radii
    f_rad : :class:`numpy.ndarray`
        Array containing the function-evaluation points within each disk

    Examples
    --------
    >>> radii(2)
    (array([   0.,  500.,   inf]), array([  0.,  inf]))
    """

    if typ == "log":
        if rinf == np.inf:
            # define the partition radii
            p_rad = np.expm1(
                np.linspace(np.log1p(rwell), np.log1p(rlast), parts)
            )
            p_rad = np.append(p_rad, [np.inf])

            # define the points within the partitions to evaluate the function
            f_rad = np.expm1(
                np.linspace(np.log1p(rwell), np.log1p(rlast), parts - 1)
            )
            f_rad = np.append(f_rad, [np.inf])
        else:
            p_rad = np.expm1(
                np.linspace(np.log1p(rwell), np.log1p(rinf), parts + 1)
            )
            f_rad = np.expm1(
                np.linspace(np.log1p(rwell), np.log1p(rinf), parts)
            )

    else:
        if rinf == np.inf:
            # define the partition radii
            p_rad = np.linspace(rwell, rlast, parts)
            p_rad = np.append(p_rad, [np.inf])

            # define the points within the partitions to evaluate the function
            f_rad = np.linspace(rwell, rlast, parts - 1)
            f_rad = np.append(f_rad, [np.inf])
        else:
            p_rad = np.linspace(rwell, rinf, parts + 1)
            f_rad = np.linspace(rwell, rinf, parts)

    return p_rad, f_rad


def specialrange(val_min, val_max, steps, typ="log"):
    """
    Calculation of special point ranges

    Parameters
    ----------
    val_min : :class:`float`
        Starting value.
    val_max : :class:`float`
        Ending value
    steps : :class:`int`
        Number of steps.
    typ : :class:`str` or :class:`float`, optional
        Setting the kind of range-distribution. One can choose between

        * ``"log"``: for logarithmic behavior
        * ``"lin"``: for linear behavior
        * ``"quad"``: for quadratic behavior
        * ``"cub"``: for cubic behavior
        * :class:`float`: here you can specifi any exponent ("quad" would be
          equivalent to 2)

        Default: ``"log"``

    Returns
    -------
    :class:`numpy.ndarray`
        Array containing the special range

    Examples
    --------
    >>> specialrange(1,10,4)
    array([  1.        ,   2.53034834,   5.23167968,  10.        ])
    """

    if typ in ["logarithmic", "log"]:
        rng = np.expm1(
            np.linspace(np.log1p(val_min), np.log1p(val_max), steps)
        )
    elif typ in ["linear", "lin"]:
        rng = np.linspace(val_min, val_max, steps)
    elif typ in ["quadratic", "quad"]:
        rng = (np.linspace(np.sqrt(val_min), np.sqrt(val_max), steps)) ** 2
    elif typ in ["cubic", "cub"]:
        rng = (
            np.linspace(
                np.power(val_min, 1 / 3.0), np.power(val_max, 1 / 3.0), steps
            )
        ) ** 3
    elif isinstance(typ, (float, int)):
        rng = (
            np.linspace(
                np.power(val_min, 1.0 / typ),
                np.power(val_max, 1.0 / typ),
                steps,
            )
        ) ** typ
    else:
        rng = np.linspace(val_min, val_max, steps)

    return rng


def specialrange_cut(val_min, val_max, steps, val_cut=np.inf, typ="log"):
    """
    Calculation of special point ranges

    Calculation of special point ranges with a cut-off value.

    Parameters
    ----------
    val_min : :class:`float`
        Starting value.
    val_max : :class:`float`
        Ending value
    steps : :class:`int`
        Number of steps.
    val_cut : :class:`float`
        Cutting value, if val_max is bigger than this value, the last interval
        will be between val_cut and val_max
    typ : :class:`str` or :class:`float`, optional
        Setting the kind of range-distribution. One can choose between

        * ``"log"``: for logarithmic behavior
        * ``"lin"``: for linear behavior
        * ``"quad"``: for quadratic behavior
        * ``"cub"``: for cubic behavior
        * :class:`float`: here you can specifi any exponent ("quad" would be
          equivalent to 2)

        Default: ``"log"``

    Returns
    -------
    :class:`numpy.ndarray`
        Array containing the special range

    Examples
    --------
    >>> specialrange_cut(1,10,4)
    array([  1.        ,   2.53034834,   5.23167968,  10.        ])
    """

    if val_max > val_cut:
        rng = specialrange(val_min, val_cut, steps - 1, typ)
        return np.hstack((rng, val_max))

    return specialrange(val_min, val_max, steps, typ)


def aniso(e):
    """
    The anisotropy function

    Known from ''Dagan (1989)''[R2]_.

    Parameters
    ----------
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths

    Returns
    -------
    aniso : :class:`float`
        Value of the anisotropy function for the given value.

    Raises
    ------
    ValueError
        If the Anisotropy-ratio ``e`` is not within 0 and 1.


    References
    ----------
    .. [R2] Dagan, G., ''Flow and Transport on Porous Formations'',
           Springer Verlag, New York, 1989.

    Examples
    --------
    >>> aniso(0.5)
    0.23639985871871511
    """

    if not 0.0 <= e <= 1.0:
        raise ValueError("Anisotropieratio 'e' must be within 0 and 1")

    if e == 1.0:
        res = 1.0 / 3.0
    elif e == 0.0:
        res = 0.0
    else:
        res = e / (2 * (1.0 - e ** 2))
        res *= (
            1.0
            / np.sqrt(1.0 - e ** 2)
            * np.arctan(np.sqrt(1.0 / e ** 2 - 1.0))
            - e
        )

    return res


def well_solution(rad, time, T, S, Qw, struc_grid=True, hinf=0.0):
    """
    The classical Theis solution

    The classical Theis solution for transient flow under a pumping condition
    in a confined and homogeneous aquifer.

    This solution was presented in ''Theis 1935''[R9]_.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    T : :class:`float`
        Given transmissivity of the aquifer
    S : :class:`float`
        Given storativity of the aquifer
    Qw : :class:`float`
        Pumpingrate at the well
    struc_grid : :class:`bool`, optional
        If this is set to "False", the "rad" and "time" array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    hinf : :class:`float`, optional
        Reference head at the outer boundary "rinf". Default: ``0.0``

    Returns
    -------
    well_solution : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Raises
    ------
    ValueError
        If ``rad`` is not positiv.
    ValueError
        If ``time`` is not positiv.
    ValueError
        If shape of ``rad`` and ``time`` differ in case of
        ``struc_grid`` is ``True``.
    ValueError
        If ``T`` is not positiv.
    ValueError
        If ``S`` is not positiv.

    References
    ----------
    .. [R9] Theis, C.,
       ''The relation between the lowering of the piezometric surface and the
       rate and duration of discharge of a well using groundwater storage'',
       Trans. Am. Geophys. Union, 16, 519-524, 1935

    Notes
    -----
    The parameters ``rad``, ``T`` and ``S`` will be checked for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> well_solution([1,2,3], [10,100], 0.001, 0.001, -0.001)
    array([[-0.24959541, -0.14506368, -0.08971485],
           [-0.43105106, -0.32132823, -0.25778313]])
    """

    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)

    if not struc_grid:
        grid_shape = rad.shape
        rad = rad.reshape(-1)

    if not np.all(rad > 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if not np.all(time > 0.0):
        raise ValueError("The given times need to be > 0")
    if not struc_grid and not rad.shape == time.shape:
        raise ValueError(
            "For unstructured grid the number of time- & radii-pts must equal"
        )
    if not T > 0.0:
        raise ValueError("The Transmissivity needs to be positiv")
    if not S > 0.0:
        raise ValueError("The Storage needs to be positiv")

    res = np.zeros(time.shape + rad.shape)

    for ti, te in np.ndenumerate(time):
        for ri, re in np.ndenumerate(rad):
            res[ti + ri] = (
                Qw / (4.0 * np.pi * T) * exp1(re ** 2 * S / (4 * T * te))
            )

    if not struc_grid and grid_shape:
        res = np.copy(np.diag(res).reshape(grid_shape))

    # add the reference head
    res += hinf

    return res


def inc_gamma(s, x):
    r"""The (upper) incomplete gamma function

    Given by: :math:`\Gamma(s,x) = \int_x^{\infty} t^{s-1}\,e^{-t}\,{\rm d}t`

    Parameters
    ----------
    s : :class:`float`
        exponent in the integral
    x : :class:`numpy.ndarray`
        input values
    """
    if np.isclose(s, 0):
        return exp1(x)
    if np.isclose(s, np.around(s)) and s < -0.5:
        return x ** (s - 1) * expn(int(1 - np.around(s)), x)
    if s < 0:
        return (inc_gamma(s + 1, x) - x ** s * np.exp(-x)) / s
    return gamma(s) * gammaincc(s, x)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing several helper functions related to coarse graining.

.. currentmodule:: anaflow.tools.coarse_graining

The following functions are provided

.. autosummary::

   T_CG
   T_CG_inverse
   T_CG_error
   K_CG
   K_CG_inverse
   K_CG_error
"""

from __future__ import absolute_import, division, print_function

import numpy as np

from anaflow.tools.special import aniso

__all__ = [
    "T_CG",
    "T_CG_inverse",
    "T_CG_error",
    "K_CG",
    "K_CG_inverse",
    "K_CG_error",
]


def T_CG(rad, TG, sig2, corr, prop=1.6, Twell=None):
    """
    The coarse-graining Transmissivity

    This solution was presented in ''Schneider & Attinger 2008''[R3]_.

    This functions gives an effective
    transmissivity for 2D pumpingtests in heterogenous aquifers, where the
    transmissivity is following a log-normal distribution and a gaussian
    correlation function.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    TG : :class:`float`
        Geometric-mean of the transmissivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the transmissivity-distribution
    corr : :class:`float`
        corralation-length of transmissivity-distribution
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    Twell : :class:`float`, optional
        Explicit transmissivity value at the well. Default: ``None``

    Returns
    -------
    T_CG : :class:`numpy.ndarray`
        Array containing the effective transmissivity values.

    References
    ----------
    .. [R3] Schneider, C. and Attinger, S.,
       ''Beyond thiem: A new method for interpreting large scale
       pumping tests in heterogeneous aquifers.''
       Water resources research, 44(4), 2008

    Examples
    --------
    >>> T_CG([1,2,3], 0.001, 1, 10, 2)
    array([ 0.00061831,  0.00064984,  0.00069236])
    """

    rad = np.squeeze(rad)

    if Twell is not None:
        chi = np.log(Twell) - np.log(TG)
    else:
        chi = -sig2 / 2.0

    return TG * np.exp(chi / (1.0 + (prop * rad / corr) ** 2))


def T_CG_inverse(T, TG, sig2, corr, prop=1.6, Twell=None):
    """
    The inverse coarse-graining Transmissivity

    See: :func:`T_CG`

    Parameters
    ----------
    T : :class:`numpy.ndarray`
        Array with all transmissivity values
        where the function should be evaluated
    TG : :class:`float`
        Geometric-mean of the transmissivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the transmissivity-distribution
    corr : :class:`float`
        corralation-length of transmissivity-distribution
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    Twell : :class:`float`, optional
        Explicit transmissivity value at the well. Default: ``None``

    Returns
    -------
    rad : :class:`numpy.ndarray`
        Array containing the radii belonging to the given transmissivity values

    Examples
    --------
    >>> T_CG_inverse([7e-4,8e-4,9e-4], 0.001, 1, 10, 2)
    array([ 3.16952925,  5.56935826,  9.67679026])
    """

    T = np.squeeze(T)

    if Twell is not None:
        chi = np.log(Twell) - np.log(TG)
    else:
        chi = -sig2 / 2.0

    return (corr / prop) * np.sqrt(chi / np.log(T / TG) - 1.0)


def T_CG_error(err, TG, sig2, corr, prop=1.6, Twell=None):
    """
    Calculating the radial-point for given error

    Calculating the radial-point where the relative error of the farfield
    value is less than the given tollerance.
    See: :func:`T_CG`

    Parameters
    ----------
    err : :class:`float`
        Given relative error for the farfield transmissivity
    TG : :class:`float`
        Geometric-mean of the transmissivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the transmissivity-distribution
    corr : :class:`float`
        corralation-length of transmissivity-distribution
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    Twell : :class:`float`, optional
        Explicit transmissivity value at the well. Default: ``None``

    Returns
    -------
    rad : :class:`float`
        Radial point, where the relative error is less than the given one.

    Examples
    --------
    >>> T_CG_error(0.01, 0.001, 1, 10, 2)
    34.910450167790387
    """

    if Twell is not None:
        chi = np.log(Twell) - np.log(TG)
    else:
        chi = -sig2 / 2.0

    if chi > 0.0:
        if chi / np.log(1.0 + err) >= 1.0:
            return (corr / prop) * np.sqrt(chi / np.log(1.0 + err) - 1.0)
        # standard value if the error is less then the variation
        return 1.0
    else:
        if chi / np.log(1.0 - err) >= 1.0:
            return (corr / prop) * np.sqrt(chi / np.log(1.0 - err) - 1.0)
        # standard value if the error is less then the variation
        return 1.0


def K_CG(rad, KG, sig2, corr, e, prop=1.6, Kwell="KH"):
    """
    The coarse-graining conductivity

    This solution was presented in ''Zech 2013''[R8]_.

    This functions gives an effective
    conductivity for 3D pumpingtests in heterogenous aquifers, where the
    conductivity is following a log-normal distribution and a gaussian
    correlation function and taking vertical anisotropy into account.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    KG : :class:`float`
        Geometric-mean conductivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the conductivity-distribution
    corr : :class:`float`
        corralation-length of conductivity-distribution
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    Kwell :  :class:`str` or  :class:`float`, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``),
        the arithmetic mean (``"KA"``) or an arbitrary float
        value. Default: ``"KH"``

    Returns
    -------
    K_CG : :class:`numpy.ndarray`
        Array containing the effective conductivity values.

    References
    ----------
    .. [R8] Zech, A.
       ''Impact of Aqifer Heterogeneity on Subsurface Flow and Salt
       Transport at Different Scales: from a method determine parameters
       of heterogeneous permeability at local scale to a large-scale model
       for the sedimentary basin of Thuringia.''
       PhD thesis, Friedrich-Schiller-Universität Jena, 2013

    Examples
    --------
    >>> K_CG([1,2,3], 0.001, 1, 10, 1, 2)
    array([ 0.00063008,  0.00069285,  0.00077595])
    """

    rad = np.squeeze(rad)

    Kefu = KG * np.exp(sig2 * (0.5 - aniso(e)))
    if Kwell == "KH":
        chi = sig2 * (aniso(e) - 1.0)
    elif Kwell == "KA":
        chi = sig2 * aniso(e)
    else:
        chi = np.log(Kwell) - np.log(Kefu)

    return Kefu * np.exp(
        chi / np.sqrt(1.0 + (prop * rad / (corr * e ** (1.0 / 3.0))) ** 2) ** 3
    )


def K_CG_inverse(K, KG, sig2, corr, e, prop=1.6, Kwell="KH"):
    """
    The inverse coarse-graining conductivity

    See: :func:`K_CG`

    Parameters
    ----------
    K : :class:`numpy.ndarray`
        Array with all conductivity values
        where the function should be evaluated
    KG : :class:`float`
        Geometric-mean conductivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the conductivity-distribution
    corr : :class:`float`
        corralation-length of conductivity-distribution
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    Kwell :  :class:`str` or  :class:`float`, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean ("KH"), the arithmetic mean ("KA") or an arbitrary float
        value. Default: ``"KH"``

    Returns
    -------
    rad : :class:`numpy.ndarray`
        Array containing the radii belonging to the given conductivity values

    Examples
    --------
    >>> K_CG_inverse([7e-4,8e-4,9e-4], 0.001, 1, 10, 1, 2)
    array([ 2.09236867,  3.27914996,  4.52143956])
    """

    K = np.squeeze(K)

    Kefu = KG * np.exp(sig2 * (0.5 - aniso(e)))
    if Kwell == "KH":
        chi = sig2 * (aniso(e) - 1.0)
    elif Kwell == "KA":
        chi = sig2 * aniso(e)
    else:
        chi = np.log(Kwell) - np.log(Kefu)

    return (
        corr
        * e ** (1.0 / 3.0)
        / prop
        * np.sqrt((chi / np.log(K / Kefu)) ** (2.0 / 3.0) - 1.0)
    )


def K_CG_error(err, KG, sig2, corr, e, prop=1.6, Kwell="KH"):
    """
    Calculating the radial-point for given error

    Calculating the radial-point where the relative error of the farfield
    value is less than the given tollerance.
    See: :func:`K_CG`

    Parameters
    ----------
    err : :class:`float`
        Given relative error for the farfield conductivity
    KG : :class:`float`
        Geometric-mean conductivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the conductivity-distribution
    corr : :class:`float`
        corralation-length of conductivity-distribution
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    Kwell :  :class:`str` or  :class:`float`, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean ("KH"), the arithmetic mean ("KA") or an arbitrary float
        value. Default: ``"KH"``

    Returns
    -------
    rad : :class:`float`
        Radial point, where the relative error is less than the given one.

    Examples
    --------
    >>> K_CG_error(0.01, 0.001, 1, 10, 1, 2)
    19.612796453639845
    """

    Kefu = KG * np.exp(sig2 * (0.5 - aniso(e)))
    if Kwell == "KH":
        chi = sig2 * (aniso(e) - 1.0)
    elif Kwell == "KA":
        chi = sig2 * aniso(e)
    else:
        chi = np.log(Kwell) - np.log(Kefu)

    coef = corr * e ** (1.0 / 3.0) / prop

    if chi > 0.0:
        if chi / np.log(1.0 + err) >= 1.0:
            return coef * np.sqrt(
                (chi / np.log(1.0 + err)) ** (2.0 / 3.0) - 1.0
            )
        # standard value if the error is less then the variation
        return 1.0
    else:
        if chi / np.log(1.0 - err) >= 1.0:
            return coef * np.sqrt(
                (chi / np.log(1.0 - err)) ** (2.0 / 3.0) - 1.0
            )
        # standard value if the error is less then the variation
        return 1.0


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing several mean calculating routines.

.. currentmodule:: anaflow.tools.mean

The following functions are provided

.. autosummary::

   rad_amean_func
   rad_gmean_func
   rad_hmean_func
   rad_pmean_func
"""

from __future__ import absolute_import, division, print_function

import numpy as np

from scipy.integrate import quad as integ

__all__ = [
    "rad_amean_func",
    "rad_gmean_func",
    "rad_hmean_func",
    "rad_pmean_func",
]


def rad_amean_func(func, val_arr, arg_dict=None, **kwargs):
    """
    Calculating the arithmetic mean

    Calculating the arithmetic mean of a radial symmetric function
    for given consecutive radii defining 2D disks by the following formula

    .. math::
       f_i=\\frac{2}{r_{i+1}^2-r_i^2}
       \\intop^{r_{i+1}}_{r_i} r\\cdot f(r)\\, dr

    Parameters
    ----------
    func : :any:`callable`
        function that should be used
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        given radii defining the disks
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``rad_amean_func`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated arithmetic means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``

    Examples
    --------
    >>> f = lambda x: x**2
    >>> rad_amean_func(f, [1,2,3])
    array([ 2.33588885,  6.33423311])
    """
    if arg_dict is None:
        arg_dict = {}
    kwargs.update(arg_dict)

    val_arr = np.array(val_arr, dtype=float).reshape(-1)
    parts = len(val_arr) - 1

    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if len(val_arr) < 2:
        raise ValueError("To few input values in val_arr. Need at least 2.")
    if not all(val_arr[i] < val_arr[i + 1] for i in range(len(val_arr) - 1)):
        raise ValueError("The input values need to be sorted")

    # dummy function defining the needed integrand
    def _step(func, kwargs):
        """
        dummy function providing the integrand
        """

        def integrand(val):
            """
            Integrand for the geometric mean ``r*log(f(r))``
            """
            return 2 * val * func(val, **kwargs)

        return integrand

    # creating the output array
    func_arr = np.zeros_like(val_arr[:-1], dtype=float)

    # iterating over the input values
    for i in range(parts):
        # if one side is infinity, the function is evaluated at infinity
        if val_arr[i + 1] == np.inf:
            func_arr[i] = func(np.inf, **kwargs)
        else:
            func_arr[i] = integ(
                _step(func, kwargs), val_arr[i], val_arr[i + 1]
            )[0]
            func_arr[i] /= val_arr[i + 1] ** 2 - val_arr[i] ** 2

    return func_arr


def rad_gmean_func(func, val_arr, arg_dict=None, **kwargs):
    """
    Calculating the geometric mean

    Calculating the geometric meanof a radial symmetric function
    for given consecutive radii defining 2D disks by the following formula

    .. math::
       f_i=\\exp\\left(\\frac{2}{r_{i+1}^2-r_i^2}
       \\intop^{r_{i+1}}_{r_i} r\\cdot\\ln(f(r))\\, dr\\right)

    Parameters
    ----------
    func : :any:`callable`
        function that should be used
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        given radii defining the disks
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``rad_gmean_func`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated geometric means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``

    Examples
    --------
    >>> f = lambda x: x**2
    >>> rad_gmean_func(f, [1,2,3])
    array([ 2.33588885,  6.33423311])
    """

    if arg_dict is None:
        arg_dict = {}
    kwargs.update(arg_dict)

    val_arr = np.array(val_arr, dtype=float).reshape(-1)
    parts = len(val_arr) - 1

    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if len(val_arr) < 2:
        raise ValueError("To few input values in val_arr. Need at least 2.")
    if not all(val_arr[i] < val_arr[i + 1] for i in range(len(val_arr) - 1)):
        raise ValueError("The input values need to be sorted")

    # dummy function defining the needed integrand
    def _step(func, kwargs):
        """
        dummy function providing the integrand
        """

        def integrand(val):
            """
            Integrand for the geometric mean ``r*log(f(r))``
            """
            return 2 * val * np.log(func(val, **kwargs))

        return integrand

    # creating the output array
    func_arr = np.zeros_like(val_arr[:-1], dtype=float)

    # iterating over the input values
    for i in range(parts):
        # if one side is infinity, the function is evaluated at infinity
        if val_arr[i + 1] == np.inf:
            func_arr[i] = func(np.inf, **kwargs)
        else:
            func_arr[i] = integ(
                _step(func, kwargs), val_arr[i], val_arr[i + 1]
            )[0]
            func_arr[i] = np.exp(
                func_arr[i] / (val_arr[i + 1] ** 2 - val_arr[i] ** 2)
            )

    return func_arr


def rad_hmean_func(func, val_arr, arg_dict=None, **kwargs):
    """
    Calculating the harmonic mean

    Calculating the harmonic mean of a radial symmetric function
    for given consecutive radii defining 2D disks by the following formula

    .. math::
       f_i=\\left(\\frac{2}{r_{i+1}^2-r_i^2}
       \\intop^{r_{i+1}}_{r_i} r\\cdot(f(r))^{-1}\\, dr\\right)^{-1}

    Parameters
    ----------
    func : :any:`callable`
        function that should be used
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        given radii defining the disks
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``rad_gmean_func`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated geometric means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``

    Examples
    --------
    >>> f = lambda x: x**2
    >>> rad_hmean_func(f, [1,2,3])
    array([ 2.33588885,  6.33423311])
    """

    if arg_dict is None:
        arg_dict = {}
    kwargs.update(arg_dict)

    val_arr = np.array(val_arr, dtype=float).reshape(-1)
    parts = len(val_arr) - 1

    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if len(val_arr) < 2:
        raise ValueError("To few input values in val_arr. Need at least 2.")
    if not all(val_arr[i] < val_arr[i + 1] for i in range(len(val_arr) - 1)):
        raise ValueError("The input values need to be sorted")

    # dummy function defining the needed integrand
    def _step(func, kwargs):
        """
        dummy function providing the integrand
        """

        def integrand(val):
            """
            Integrand for the geometric mean ``r*log(f(r))``
            """
            return 2 * val / func(val, **kwargs)

        return integrand

    # creating the output array
    func_arr = np.zeros_like(val_arr[:-1], dtype=float)

    # iterating over the input values
    for i in range(parts):
        # if one side is infinity, the function is evaluated at infinity
        if val_arr[i + 1] == np.inf:
            func_arr[i] = func(np.inf, **kwargs)
        else:
            func_arr[i] = integ(
                _step(func, kwargs), val_arr[i], val_arr[i + 1]
            )[0]
            func_arr[i] = 1.0 / (
                func_arr[i] / (val_arr[i + 1] ** 2 - val_arr[i] ** 2)
            )

    return func_arr


def rad_pmean_func(func, val_arr, p=1.0, arg_dict=None, **kwargs):
    """
    Calculating the p-mean

    Calculating the p-mean of a radial symmetric function
    for given consecutive radii defining 2D disks by the following formula

    .. math::
       f_i=\\left(\\frac{2}{r_{i+1}^2-r_i^2}
       \\intop^{r_{i+1}}_{r_i} r\\cdot(f(r))^p\\, dr\\right)^{\\frac{1}{p}}

    Parameters
    ----------
    func : :any:`callable`
        function that should be used
        The first argument needs to be the radial variable:
        ``func(r, **kwargs)``
    val_arr : :class:`numpy.ndarray`
        given radii defining the disks
    p : :class:`float`, optional
        The potency defining the p-mean.
        Default: ``1.0``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``.
        This is designed for overlapping keywords in ``rad_pmean_func`` and
        ``func``.
        Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all calculated p-means

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If ``val_arr`` has less than 2 values.
    ValueError
        If ``val_arr`` is not sorted in incresing order.

    Notes
    -----
    If the last value in val_arr is "inf", the given function should provide
    a value for "inf" as input: ``func(float("inf"))``

    Examples
    --------
    >>> f = lambda x: x**2
    >>> rad_pmean_func(f, [1,2,3])
    array([ 2.33588885,  6.33423311])
    """

    # if p is 0, the limit-case of the geometric mean is returned
    if p == 0:
        return rad_gmean_func(func, val_arr, arg_dict, **kwargs)

    if arg_dict is None:
        arg_dict = {}
    kwargs.update(arg_dict)

    val_arr = np.array(val_arr, dtype=float).reshape(-1)
    parts = len(val_arr) - 1

    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if len(val_arr) < 2:
        raise ValueError("To few input values in val_arr. Need at least 2.")
    if not all(val_arr[i] < val_arr[i + 1] for i in range(len(val_arr) - 1)):
        raise ValueError("The input values need to be sorted")

    # dummy function defining the needed integrand
    def _step(func, kwargs):
        """
        dummy function providing the integrand
        """

        def integrand(val):
            """
            Integrand for the geometric mean ``r*log(f(r))``
            """
            return 2 * val * func(val, **kwargs) ** p

        return integrand

    # creating the output array
    func_arr = np.zeros_like(val_arr[:-1], dtype=float)

    # iterating over the input values
    for i in range(parts):
        # if one side is infinity, the function is evaluated at infinity
        if val_arr[i + 1] == np.inf:
            func_arr[i] = func(np.inf, **kwargs)
        else:
            func_arr[i] = integ(
                _step(func, kwargs), val_arr[i], val_arr[i + 1]
            )[0]
            func_arr[i] = (
                func_arr[i] / (val_arr[i + 1] ** 2 - val_arr[i] ** 2)
            ) ** (1.0 / p)

    return func_arr


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
""" AnaFlow - analytical solutions for the groundwater-flow equation"""

import os
import codecs
import re

from setuptools import setup, find_packages


# find __version__ ############################################################


def read(*parts):
    """read file data"""
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    """find version without importing module"""
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


###############################################################################


DOCLINES = __doc__.split("\n")
README = open("README.md").read()
CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Developers
Intended Audience :: End Users/Desktop
Intended Audience :: Science/Research
License :: OSI Approved :: GNU General Public License v3 (GPLv3)
Natural Language :: English
Operating System :: MacOS
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
Programming Language :: Python
Programming Language :: Python :: 2
Programming Language :: Python :: 3
Topic :: Scientific/Engineering
Topic :: Software Development
Topic :: Utilities
"""

VERSION = find_version("anaflow", "_version.py")

setup(
    name='anaflow',
    version=VERSION,
    maintainer="Sebastian Mueller",
    maintainer_email="sebastian.mueller@ufz.de",
    description=DOCLINES[0],
    long_description=README,
    long_description_content_type="text/markdown",
    author="Sebastian Mueller",
    author_email="sebastian.mueller@ufz.de",
    url='https://github.com/GeoStat-Framework/AnaFlow',
    license='GPL - see LICENSE',
    classifiers=[_f for _f in CLASSIFIERS.split('\n') if _f],
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    include_package_data=True,
    install_requires=['numpy >= 1.13.0',
                      'scipy >= 0.19.1'],
    packages=find_packages(exclude=['tests*', 'docs*']),
)



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing functions concerning the laplace transformation.

.. currentmodule:: anaflow.tools.laplace

The following functions are provided

.. autosummary::

   get_lap
   get_lap_inv
   lap_trans
   stehfest
"""

from __future__ import absolute_import, division, print_function

from math import floor, factorial
import numpy as np
from scipy.integrate import quad

__all__ = ["get_lap", "lap_trans", "get_lap_inv", "stehfest"]


def get_lap(func, arg_dict=None, **kwargs):
    """
    Callable Laplace transform

    Get the Laplace transform of a given function as a callable function.

    Parameters
    ----------
    func : :any:`callable`
        function that shall be transformed.
        The first argument needs to be the time-variable:
        ``func(t, **kwargs)``

        `func` should be capable of taking numpy arrays as input for `s` and
        the first shape component of the output of `func` should match the
        shape of `s`.
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``
        This is designed for overlapping keywords. Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``.

    Returns
    -------
    :any:`callable`
        The Laplace transformed of the given function.

    Raises
    ------
    ValueError
        If `func` is not callable.
    """
    # check the input
    if not callable(func):
        raise ValueError("The given function needs to be callable")

    # define the returned function
    def ret_func(phase):
        """Return function for the Laplace transformed."""
        return lap_trans(func, phase, arg_dict=arg_dict, **kwargs)

    return ret_func


def lap_trans(func, phase, arg_dict=None, **kwargs):
    """
    The laplace transform

    Parameters
    ----------
    func : :any:`callable`
        function that shall be transformed.
        The first argument needs to be the time-variable:
        ``func(s, **kwargs)``

        `func` should be capable of taking numpy arrays as input for `s` and
        the first shape component of the output of `func` should match the
        shape of `s`.
    phase : :class:`float` or :class:`numpy.ndarray`
        phase-points to evaluate the transformed function at
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``
        This is designed for overlapping keywords in ``stehfest`` and
        ``func``.Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all evaluations in phase-space.

    Raises
    ------
    ValueError
        If `func` is not callable.
    """
    # check the input
    if not callable(func):
        raise ValueError("The given function needs to be callable")

    if arg_dict is None:
        arg_dict = {}
    kwargs.update(arg_dict)

    # check and save if phase is scalar
    is_scal = np.isscalar(phase)

    # ensure that t is handled as an 1d-array
    phase = np.array(phase, dtype=float)
    result = np.zeros_like(phase)

    def make_integrand(ph):
        def integrand(val):
            """
            Integrand for the laplace transform
            """
            return np.exp(-ph * val) * func(val, **kwargs)

        return integrand

    for phase_i, phase_e in np.ndenumerate(phase):

        integ = make_integrand(phase_e)
        result[phase_i] = quad(integ, 0, np.inf)[0]

    if is_scal:
        result = np.asscalar(result)

    return result


def get_lap_inv(
    func, method="stehfest", method_dict=None, arg_dict=None, **kwargs
):
    """
    Callable Laplace inversion

    Get the Laplace inversion of a given function as a callable function.

    Parameters
    ----------
    func : :any:`callable`
        function in laplace-space that shall be inverted.
        The first argument needs to be the laplace-variable:
        ``func(s, **kwargs)``

        `func` should be capable of taking numpy arrays as input for `s` and
        the first shape component of the output of `func` should match the
        shape of `s`.
    method : :class:`str`
        Method that should be used to calculate the inverse.
        One can choose between

        * ``"stehfest"``: for the stehfest algorithm

        Default: ``"stehfest"``
    method_dict : :class:`dict` or :any:`None`, optional
        Keyword arguments for the used method.
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``
        This is designed for overlapping keywords. Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``.

    Returns
    -------
    :any:`callable`
        The Laplace inverse of the given function.

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If `method` is unknown.
    """
    # dict with all implemented methods
    method_avail = {"stehfest": stehfest}
    # check the input
    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if method not in method_avail:
        raise ValueError("The given method is unknown: " + str(method))
    # assign the used method
    used_meth = method_avail[method]
    # update kwargs
    if method_dict is None:
        method_dict = {}
    kwargs.update(method_dict)
    kwargs["arg_dict"] = arg_dict

    # define the returned function
    def ret_func(time):
        """Return function for the Laplace inversion."""
        return used_meth(func, time, **kwargs)

    return ret_func


def stehfest(func, time, bound=12, arg_dict=None, **kwargs):
    """
    The stehfest-algorithm for numerical laplace inversion

    The Inversion was derivide in ''Stehfest 1970''[R1]_
    and is given by the formula

    .. math::
       f\\left(t\\right) &=\\frac{\\ln2}{t}\\sum_{n=1}^{N}c_{n}\\cdot\\tilde{f}
       \\left(n\\cdot\\frac{\\ln2}{t}\\right)\\\\
       c_{n} &=\\left(-1\\right)^{n+\\frac{N}{2}}\\cdot
       \\sum_{k=\\left\\lfloor \\frac{n+1}{2}\\right\\rfloor }
       ^{\\min\\left\\{ n,\\frac{N}{2}\\right\\} }
       \\frac{k^{\\frac{N}{2}+1}\\cdot\\binom{2k}{k}}
       {\\left(\\frac{N}{2}-k\\right)!\\cdot\\left(n-k\\right)!
       \\cdot\\left(2k-n\\right)!}

    In the algorithm
    :math:`N` corresponds to ``bound``,
    :math:`\\tilde{f}` to ``func`` and
    :math:`t` to ``time``.

    Parameters
    ----------
    func : :any:`callable`
        function in laplace-space that shall be inverted.
        The first argument needs to be the laplace-variable:
        ``func(s, **kwargs)``

        `func` should be capable of taking numpy arrays as input for `s` and
        the first shape component of the output of `func` should match the
        shape of `s`.
    time : :class:`float` or :class:`numpy.ndarray`
        time-points to evaluate the function at
    bound : :class:`int`, optional
        Here you can specify the number of interations within this
        algorithm. Default: ``12``
    arg_dict : :class:`dict` or :any:`None`, optional
        Keyword-arguments given as a dictionary that are forwarded to the
        function given in ``func``. Will be merged with ``**kwargs``
        This is designed for overlapping keywords in ``stehfest`` and
        ``func``.Default: ``None``
    **kwargs
        Keyword-arguments that are forwarded to the function given in ``func``.
        Will be merged with ``arg_dict``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all evaluations in Time-space.

    Raises
    ------
    ValueError
        If `func` is not callable.
    ValueError
        If `time` is not positive.
    ValueError
        If `bound` is not positive.
    ValueError
        If `bound` is not even.

    References
    ----------
    .. [R1] Stehfest, H., ''Algorithm 368:
       Numerical inversion of laplace transforms [d5].''
       Communications of the ACM, 13(1):47-49, 1970

    Notes
    -----
    The parameter ``time`` needs to be strictly positiv.

    The algorithm gets unstable for ``bound`` values above 20.

    Examples
    --------
    >>> f = lambda x: x**-1
    >>> stehfest(f, [1,10,100])
    array([ 1.,  1.,  1.])
    """

    if arg_dict is None:
        arg_dict = {}
    kwargs.update(arg_dict)

    # check and save if 't' is scalar
    is_scal = np.isscalar(time)

    # ensure that t is handled as an 1d-array
    time = np.array(time, dtype=float).reshape(-1)

    # check the input
    if not callable(func):
        raise ValueError("The given function needs to be callable")
    if not np.all(time > 0.0):
        raise ValueError(
            "The time-values need to be positiv for the stehfest-algorithm"
        )
    if bound <= 1:
        raise ValueError(
            "The boundary needs to be >1 for the stehfest-algorithm"
        )
    if bound % 2 != 0:
        raise ValueError(
            "The boundary needs to be even for the stehfest-algorithm"
        )

    # get all coefficient factors at once
    c_fac = c_array(bound)
    t_fac = np.log(2.0) / time

    # store every function-argument needed in one array
    fargs = np.outer(t_fac, np.arange(1, bound + 1))

    # get every function-value needed with one call of 'func'
    lap_val = func(fargs.reshape(-1), **kwargs)
    lap_val = lap_val.reshape(fargs.shape + lap_val.shape[1:])

    # do all the sumation with fancy indexing in numpy
    res = np.tensordot(lap_val, c_fac, axes=(1, 0))
    res = np.rollaxis(np.multiply(np.rollaxis(res, 0, res.ndim), t_fac), -1, 0)

    # reformat the result according to the input
    res = np.squeeze(res)
    if np.ndim(res) == 0 and is_scal:
        res = np.asscalar(res)

    return res


def c_array(bound=12):
    """
    Array of coefficients for the stehfest-algorithm.

    Parameters
    ----------
    bound : :class:`int`, optional
        The number of interations. Default: ``12``

    Returns
    -------
    :class:`numpy.ndarray`
        Array with all coefficinets needed.
    """
    c_lookup = {
        2: np.array([2.000000000000000000e00, -2.000000000000000000e00]),
        4: np.array(
            [
                -2.000000000000000000e00,
                2.600000000000000000e01,
                -4.800000000000000000e01,
                2.400000000000000000e01,
            ]
        ),
        6: np.array(
            [
                1.000000000000000000e00,
                -4.900000000000000000e01,
                3.660000000000000000e02,
                -8.580000000000000000e02,
                8.100000000000000000e02,
                -2.700000000000000000e02,
            ]
        ),
        8: np.array(
            [
                -3.333333333333333148e-01,
                4.833333333333333570e01,
                -9.060000000000000000e02,
                5.464666666666666060e03,
                -1.437666666666666606e04,
                1.873000000000000000e04,
                -1.194666666666666606e04,
                2.986666666666666515e03,
            ]
        ),
        10: np.array(
            [
                8.333333333333332871e-02,
                -3.208333333333333570e01,
                1.279000000000000000e03,
                -1.562366666666666606e04,
                8.424416666666665697e04,
                -2.369575000000000000e05,
                3.759116666666666861e05,
                -3.400716666666666861e05,
                1.640625000000000000e05,
                -3.281250000000000000e04,
            ]
        ),
        12: np.array(
            [
                -1.666666666666666644e-02,
                1.601666666666666572e01,
                -1.247000000000000000e03,
                2.755433333333333212e04,
                -2.632808333333333139e05,
                1.324138699999999953e06,
                -3.891705533333333209e06,
                7.053286333333333023e06,
                -8.005336500000000000e06,
                5.552830500000000000e06,
                -2.155507200000000186e06,
                3.592512000000000116e05,
            ]
        ),
        14: np.array(
            [
                2.777777777777777884e-03,
                -6.402777777777778567e00,
                9.240499999999999545e02,
                -3.459792777777777519e04,
                5.403211111111111240e05,
                -4.398346366666667163e06,
                2.108759177777777612e07,
                -6.394491304444444180e07,
                1.275975795499999970e08,
                -1.701371880833333433e08,
                1.503274670333333313e08,
                -8.459216150000000000e07,
                2.747888476666666567e07,
                -3.925554966666666791e06,
            ]
        ),
        16: np.array(
            [
                -3.968253968253968251e-04,
                2.133730158730158699e00,
                -5.510166666666666515e02,
                3.350016111111111240e04,
                -8.126651111111111240e05,
                1.007618376666666567e07,
                -7.324138297777777910e07,
                3.390596320730158687e08,
                -1.052539536278571367e09,
                2.259013328583333492e09,
                -3.399701984433333397e09,
                3.582450461699999809e09,
                -2.591494081366666794e09,
                1.227049828766666651e09,
                -3.427345554285714030e08,
                4.284181942857142538e07,
            ]
        ),
    }
    if bound in c_lookup:
        return c_lookup[bound]
    return _carr(bound)


def _carr(bound):
    res = np.zeros(bound)
    for i in range(1, bound + 1):
        res[i - 1] = _c(i, bound)
    return res


def _c(i, bound):
    res = 0.0
    for k in range(int(floor((i + 1) / 2.0)), min(i, bound // 2) + 1):
        res += _d(k, i, bound)
    res *= (-1) ** (i + bound / 2)
    return res


def _d(k, i, bound):
    res = ((float(k)) ** (bound / 2 + 1)) * (factorial(2 * k))
    res /= (
        (factorial(bound / 2 - k))
        * (factorial(i - k))
        * (factorial(2 * k - i))
    )
    res /= factorial(k) ** 2
    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing flow solutions in heterogeneous aquifers.

.. currentmodule:: anaflow.flow.heterogeneous

The following functions are provided

.. autosummary::
   ext_thiem2D
   ext_thiem3D
   ext_theis2D
   ext_theis3D
"""
from __future__ import absolute_import, division, print_function

import numpy as np
from scipy.special import exp1, expi

from anaflow.tools.laplace import stehfest as sf
from anaflow.tools.special import aniso, specialrange_cut
from anaflow.tools.mean import rad_hmean_func
from anaflow.tools.coarse_graining import T_CG, T_CG_error, K_CG, K_CG_error
from anaflow.flow.laplace import grf_laplace

__all__ = ["ext_thiem2D", "ext_thiem3D", "ext_theis2D", "ext_theis3D"]


###############################################################################
# 2D version of extended Thiem
###############################################################################


def ext_thiem2D(rad, Rref, TG, sig2, corr, Qw, href=0.0, Twell=None, prop=1.6):
    """
    The extended Thiem solution in 2D

    The extended Thiem solution for steady-state flow under
    a pumping condition in a confined aquifer.
    This solution was presented in ''Zech 2013''[R4]_.
    The type curve is describing the effective drawdown
    in a 2D statistical framework, where the transmissivity distribution is
    following a log-normal distribution with a gaussian correlation function.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    Rref : :class:`float`
        Reference radius with known head (see `href`)
    TG : :class:`float`
        Geometric-mean transmissivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the transmissivity-distribution
    corr : :class:`float`
        corralation-length of transmissivity-distribution
    Qw : :class:`float`
        Pumpingrate at the well
    href : :class:`float`, optional
        Reference head at the reference-radius `Rref`. Default: ``0.0``
    Twell : :class:`float`, optional
        Explicit transmissivity value at the well. Default: ``None``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    ext_thiem2D : :class:`numpy.ndarray`
        Array with all heads at the given radii.

    References
    ----------
    .. [R4] Zech, A.
       ''Impact of Aqifer Heterogeneity on Subsurface Flow and Salt
       Transport at Different Scales: from a method determine parameters
       of heterogeneous permeability at local scale to a large-scale model
       for the sedimentary basin of Thuringia.''
       PhD thesis, Friedrich-Schiller-Universität Jena, 2013

    Notes
    -----
    The parameters ``rad``, ``Rref``, ``TG``, ``sig2``, ``corr``, ``Twell``
    and ``prop`` will be checked for positivity.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> ext_thiem2D([1,2,3], 10, 0.001, 1, 10, -0.001)
    array([-0.53084596, -0.35363029, -0.25419375])
    """

    rad = np.squeeze(rad)

    # check the input
    if Rref <= 0.0:
        raise ValueError(
            "The upper boundary needs to be greater than the wellradius"
        )
    if np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if TG <= 0.0:
        raise ValueError("The Transmissivity needs to be positiv")
    if Twell is not None and Twell <= 0.0:
        raise ValueError("The Transmissivity at the well needs to be positiv")
    if sig2 <= 0.0:
        raise ValueError("The variance needs to be positiv")
    if corr <= 0.0:
        raise ValueError("The correlationlength needs to be positiv")
    if prop <= 0.0:
        raise ValueError("The proportionalityfactor needs to be positiv")

    # define some substitions to shorten the result
    if Twell is not None:
        chi = np.log(Twell) - np.log(TG)
    else:
        chi = -sig2 / 2.0

    Q = -Qw / (4.0 * np.pi * TG)
    C = (prop / corr) ** 2

    # derive the result
    res = -expi(-chi / (1.0 + C * rad ** 2))
    res -= np.exp(-chi) * exp1(chi / (1.0 + C * rad ** 2) - chi)
    res += expi(-chi / (1.0 + C * Rref ** 2))
    res += np.exp(-chi) * exp1(chi / (1.0 + C * Rref ** 2) - chi)
    res *= Q
    res += href

    return res


###############################################################################
# 3D version of extended Thiem
###############################################################################


def ext_thiem3D(
    rad, Rref, KG, sig2, corr, e, Qw, L, href=0.0, Kwell="KH", prop=1.6
):
    """
    The extended Thiem solution in 3D

    The extended Thiem solution for steady-state flow under
    a pumping condition in a confined aquifer.
    This solution was presented in ''Zech 2013''[R5]_.
    The type curve is describing the effective drawdown
    in a 3D statistical framework, where the conductivity distribution is
    following a log-normal distribution with a gaussian correlation function
    and taking vertical anisotropy into account.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    Rref : :class:`float`
        Reference radius with known head (see `href`)
    KG : :class:`float`
        Geometric-mean conductivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the conductivity-distribution
    corr : :class:`float`
        corralation-length of conductivity-distribution
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths
    Qw : :class:`float`
        Pumpingrate at the well
    L : :class:`float`
        Thickness of the aquifer
    href : :class:`float`, optional
        Reference head at the reference-radius `Rref`. Default: ``0.0``
    Kwell : string/float, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``), the arithmetic mean (``"KA"``) or an
        arbitrary float value. Default: ``"KH"``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``

    Returns
    -------
    ext_thiem3D : :class:`numpy.ndarray`
        Array with all heads at the given radii.

    References
    ----------
    .. [R5] Zech, A.
       ''Impact of Aqifer Heterogeneity on Subsurface Flow and Salt
       Transport at Different Scales: from a method determine parameters
       of heterogeneous permeability at local scale to a large-scale model
       for the sedimentary basin of Thuringia.''
       PhD thesis, Friedrich-Schiller-Universität Jena, 2013

    Notes
    -----
    The parameters ``rad``, ``Rref``, ``KG``, ``sig2``, ``corr``, ``Kwell``
    and ``prop`` will be checked for positivity.
    The anisotropy ``e`` factor must be greater 0 and less or equal 1.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> ext_thiem3D([1,2,3], 10, 0.001, 1, 10, 1, -0.001, 1)
    array([-0.48828026, -0.31472059, -0.22043022])
    """

    rad = np.squeeze(rad)

    # check the input
    if Rref <= 0.0:
        raise ValueError(
            "The upper boundary needs to be greater than the wellradius"
        )
    if np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if Kwell != "KA" and Kwell != "KH" and not isinstance(Kwell, float):
        raise ValueError(
            "The well-conductivity should be given as float or 'KA' resp 'KH'"
        )
    if isinstance(Kwell, float) and Kwell <= 0.0:
        raise ValueError("The well-conductivity needs to be positiv")
    if KG <= 0.0:
        raise ValueError("The Transmissivity needs to be positiv")
    if sig2 <= 0.0:
        raise ValueError("The variance needs to be positiv")
    if corr <= 0.0:
        raise ValueError("The correlationlength needs to be positiv")
    if L <= 0.0:
        raise ValueError("The aquifer-thickness needs to be positiv")
    if not 0.0 < e <= 1.0:
        raise ValueError("The anisotropy-ratio must be > 0 and <= 1")
    if prop <= 0.0:
        raise ValueError("The proportionalityfactor needs to be positiv")

    # define some substitions to shorten the result
    Kefu = KG * np.exp(sig2 * (0.5 - aniso(e)))
    if Kwell == "KH":
        chi = sig2 * (aniso(e) - 1.0)
    elif Kwell == "KA":
        chi = sig2 * aniso(e)
    else:
        chi = np.log(Kwell) - np.log(Kefu)

    Q = -Qw / (2.0 * np.pi * Kefu)
    C = (prop / corr / e ** (1.0 / 3.0)) ** 2

    sub11 = np.sqrt(1.0 + C * Rref ** 2)
    sub12 = np.sqrt(1.0 + C * rad ** 2)

    sub21 = np.log(sub12 + 1.0) - np.log(sub11 + 1.0)
    sub21 -= 1.0 / sub12 - 1.0 / sub11

    sub22 = np.log(sub12) - np.log(sub11)
    sub22 -= 0.50 / sub12 ** 2 - 0.50 / sub11 ** 2
    sub22 -= 0.25 / sub12 ** 4 - 0.25 / sub11 ** 4

    # derive the result
    res = np.exp(-chi) * (np.log(rad) - np.log(Rref))
    res += sub21 * np.sinh(chi) + sub22 * (1.0 - np.cosh(chi))
    res *= Q
    res += href

    return res


###############################################################################
# 2D version of extended Theis
###############################################################################


def ext_theis2D(
    rad,
    time,
    TG,
    sig2,
    corr,
    S,
    Qw,
    struc_grid=True,
    rwell=0.0,
    rinf=np.inf,
    hinf=0.0,
    Twell=None,
    T_err=0.01,
    prop=1.6,
    stehfestn=12,
    parts=30,
):
    """
    The extended Theis solution in 2D

    The extended Theis solution for transient flow under
    a pumping condition in a confined aquifer.
    The type curve is describing the effective drawdown
    in a 2D statistical framework, where the transmissivity distribution is
    following a log-normal distribution with a gaussian correlation function.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    TG : :class:`float`
        Geometric-mean transmissivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the transmissivity-distribution
    corr : :class:`float`
        corralation-length of transmissivity-distribution
    S : :class:`float`
        Given storativity of the aquifer
    Qw : :class:`float`
        Pumpingrate at the well
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    rwell : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    rinf : :class:`float`, optional
        Radius of the outer boundary of the aquifer. Default: ``np.inf``
    hinf : :class:`float`, optional
        Reference head at the outer boundary ``rinf``. Default: ``0.0``
    Twell : :class:`float`, optional
        Explicit transmissivity value at the well. Default: ``None``
    T_err : :class:`float`, optional
        Absolute error for the farfield transmissivity for calculating the
        cutoff-point of the solution. Default: ``0.01``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    stehfestn : :class:`int`, optional
        Since the solution is calculated in Laplace-space, the
        back-transformation is performed with the stehfest-algorithm.
        Here you can specify the number of interations within this
        algorithm. Default: ``12``
    parts : :class:`int`, optional
        Since the solution is calculated by setting the transmissity to local
        constant values, one needs to specify the number of partitions of the
        transmissivity. Default: ``30``

    Returns
    -------
    ext_theis2D : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Notes
    -----
    The parameters ``rad``, ``Rref``, ``TG``, ``sig2``, ``corr``, ``S``,
    ``Twell`` and ``prop`` will be checked for positivity.
    ``T_err`` must be greater 0 and less or equal 1.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> ext_theis2D([1,2,3], [10,100], 0.001, 1, 10, 0.001, -0.001)
    array([[-0.3381231 , -0.17430066, -0.09492601],
           [-0.58557452, -0.40907021, -0.31112835]])
    """

    # ensure that 'rad' and 'time' are arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)

    if not struc_grid:
        grid_shape = rad.shape
        rad = rad.reshape(-1)

    # check the input
    if rwell < 0.0:
        raise ValueError("The wellradius needs to be >= 0")
    if rinf <= rwell:
        raise ValueError(
            "The upper boundary needs to be greater than the wellradius"
        )
    if np.any(rad < rwell) or np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if np.any(time <= 0.0):
        raise ValueError("The given times need to be >= 0")
    if not struc_grid and not rad.shape == time.shape:
        raise ValueError(
            "For unstructured grid the number of time- & radii-pts must equal"
        )
    if TG <= 0.0:
        raise ValueError("The Transmissivity needs to be positiv")
    if Twell is not None and Twell <= 0.0:
        raise ValueError("The Transmissivity at the well needs to be positiv")
    if sig2 <= 0.0:
        raise ValueError("The variance needs to be positiv")
    if corr <= 0.0:
        raise ValueError("The correlationlength needs to be positiv")
    if S <= 0.0:
        raise ValueError("The Storage needs to be positiv")
    if prop <= 0.0:
        raise ValueError("The proportionalityfactor needs to be positiv")
    if not isinstance(stehfestn, int):
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be an integer"
        )
    if stehfestn <= 1:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be > 1"
        )
    if stehfestn % 2 != 0:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be even"
        )
    if not isinstance(parts, int):
        raise ValueError("The numbor of partitions needs to be an integer")
    if parts <= 1:
        raise ValueError("The numbor of partitions needs to be at least 2")
    if not 0.0 < T_err < 1.0:
        raise ValueError(
            "The relative error of Transmissivity needs to be within (0,1)"
        )

    # genearte rlast from a given relativ-error to farfield-transmissivity
    rlast = T_CG_error(T_err, TG, sig2, corr, prop, Twell)

    # generate the partition points
    rpart = specialrange_cut(rwell, rinf, parts + 1, rlast)

    # calculate the harmonic mean transmissivity values within each partition
    Tpart = rad_hmean_func(
        T_CG, rpart, TG=TG, sig2=sig2, corr=corr, prop=prop, Twell=Twell
    )

    # write the paramters in kwargs to use the stehfest-algorithm
    kwargs = {
        "rad": rad,
        "Qw": Qw,
        "dim": 2,
        "lat_ext": 1,
        "rpart": rpart,
        "Spart": S * np.ones(parts),
        "Kpart": Tpart,
        "Kwell": T_CG(rwell, TG, sig2, corr, prop, Twell),
    }

    # call the stehfest-algorithm
    res = sf(grf_laplace, time, bound=stehfestn, **kwargs)

    # if the input are unstructured space-time points, return an array
    if not struc_grid and grid_shape:
        res = np.copy(np.diag(res).reshape(grid_shape))

    # add the reference head
    res += hinf

    return res


###############################################################################
# 3D version of extended Theis
###############################################################################


def ext_theis3D(
    rad,
    time,
    KG,
    sig2,
    corr,
    e,
    S,
    Qw,
    L,
    struc_grid=True,
    rwell=0.0,
    rinf=np.inf,
    hinf=0.0,
    Kwell="KH",
    K_err=0.01,
    prop=1.6,
    stehfestn=12,
    parts=30,
):
    """
    The extended Theis solution in 3D

    The extended Theis solution for transient flow under
    a pumping condition in a confined aquifer.
    The type curve is describing the effective drawdown
    in a 3D statistical framework, where the transmissivity distribution is
    following a log-normal distribution with a gaussian correlation function
    and taking vertical anisotropy into account.

    Parameters
    ----------
    rad : :class:`numpy.ndarray`
        Array with all radii where the function should be evaluated
    time : :class:`numpy.ndarray`
        Array with all time-points where the function should be evaluated
    KG : :class:`float`
        Geometric-mean conductivity-distribution
    sig2 : :class:`float`
        log-normal-variance of the conductivity-distribution
    corr : :class:`float`
        corralation-length of conductivity-distribution
    e : :class:`float`
        Anisotropy-ratio of the vertical and horizontal corralation-lengths
    S : :class:`float`
        Given storativity of the aquifer
    Qw : :class:`float`
        Pumpingrate at the well
    L : :class:`float`
        Thickness of the aquifer
    struc_grid : :class:`bool`, optional
        If this is set to ``False``, the `rad` and `time` array will be merged
        and interpreted as single, r-t points. In this case they need to have
        the same shapes. Otherwise a structured r-t grid is created.
        Default: ``True``
    rwell : :class:`float`, optional
        Inner radius of the pumping-well. Default: ``0.0``
    rinf : :class:`float`, optional
        Radius of the outer boundary of the aquifer. Default: ``np.inf``
    hinf : :class:`float`, optional
        Reference head at the outer boundary `rinf`. Default: ``0.0``
    Kwell : string/float, optional
        Explicit conductivity value at the well. One can choose between the
        harmonic mean (``"KH"``), the arithmetic mean (``"KA"``) or an
        arbitrary float value. Default: ``"KH"``
    K_err : :class:`float`, optional
        Absolute error for the farfield conductivity for calculating the
        cutoff-point of the solution, if ``rinf=inf``. Default: ``0.01``
    prop: :class:`float`, optional
        Proportionality factor used within the upscaling procedure.
        Default: ``1.6``
    stehfestn : :class:`int`, optional
        Since the solution is calculated in Laplace-space, the
        back-transformation is performed with the stehfest-algorithm.
        Here you can specify the number of interations within this
        algorithm. Default: ``12``
    parts : :class:`int`, optional
        Since the solution is calculated by setting the transmissity to local
        constant values, one needs to specify the number of partitions of the
        transmissivity. Default: ``30``

    Returns
    -------
    ext_theis3D : :class:`numpy.ndarray`
        Array with all heads at the given radii and time-points.

    Notes
    -----
    The parameters ``rad``, ``Rref``, ``KG``, ``sig2``, ``corr``, ``S``,
    ``Twell`` and ``prop`` will be checked for positivity.
    The Anisotropy-ratio ``e`` must be greater 0 and less or equal 1.
    ``T_err`` must be greater 0 and less or equal 1.
    If you want to use cartesian coordiantes, just use the formula
    ``r = sqrt(x**2 + y**2)``

    Examples
    --------
    >>> ext_theis3D([1,2,3], [10,100], 0.001, 1, 10, 1, 0.001, -0.001, 1)
    array([[-0.32845576, -0.16741654, -0.09134294],
           [-0.54238241, -0.36982686, -0.27754856]])
    """

    # ensure that 'rad' and 'time' are arrays
    rad = np.squeeze(rad)
    time = np.array(time).reshape(-1)

    if not struc_grid:
        grid_shape = rad.shape
        rad = rad.reshape(-1)

    # check the input
    if rwell < 0.0:
        raise ValueError("The wellradius needs to be >= 0")
    if rinf <= rwell:
        raise ValueError(
            "The upper boundary needs to be greater than the wellradius"
        )
    if np.any(rad < rwell) or np.any(rad <= 0.0):
        raise ValueError(
            "The given radii need to be greater than the wellradius"
        )
    if np.any(time <= 0.0):
        raise ValueError("The given times need to be >= 0")
    if not struc_grid and not rad.shape == time.shape:
        raise ValueError(
            "For unstructured grid the number of time- & radii-pts must equal"
        )
    if Kwell != "KA" and Kwell != "KH" and not isinstance(Kwell, float):
        raise ValueError(
            "The well-conductivity should be given as float or 'KA' resp 'KH'"
        )
    if isinstance(Kwell, float) and Kwell <= 0.0:
        raise ValueError("The well-conductivity needs to be positiv")
    if KG <= 0.0:
        raise ValueError("The conductivity needs to be positiv")
    if sig2 <= 0.0:
        raise ValueError("The variance needs to be positiv")
    if corr <= 0.0:
        raise ValueError("The correlationlength needs to be positiv")
    if S <= 0.0:
        raise ValueError("The Storage needs to be positiv")
    if L <= 0.0:
        raise ValueError("The aquifer-thickness needs to be positiv")
    if prop <= 0.0:
        raise ValueError("The proportionalityfactor needs to be positiv")
    if not isinstance(stehfestn, int):
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be an integer"
        )
    if stehfestn <= 1:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be > 1"
        )
    if stehfestn % 2 != 0:
        raise ValueError(
            "The boundary for the Stehfest-algorithm needs to be even"
        )
    if not isinstance(parts, int):
        raise ValueError("The numbor of partitions needs to be an integer")
    if parts <= 1:
        raise ValueError("The numbor of partitions needs to be at least 2")
    if not 0.0 < K_err < 1.0:
        raise ValueError(
            "The relative error of Transmissivity needs to be within (0,1)"
        )

    # genearte rlast from a given relativ-error to farfield-conductivity
    rlast = K_CG_error(K_err, KG, sig2, corr, e, prop, Kwell=Kwell)

    # generate the partition points
    rpart = specialrange_cut(rwell, rinf, parts + 1, rlast)

    # calculate the harmonic mean conductivity values within each partition
    Kpart = rad_hmean_func(
        K_CG, rpart, KG=KG, sig2=sig2, corr=corr, e=e, prop=prop, Kwell=Kwell
    )

    # write the paramters in kwargs to use the stehfest-algorithm
    kwargs = {
        "rad": rad,
        "Qw": Qw,
        "dim": 2,
        "lat_ext": L,
        "rpart": rpart,
        "Spart": S * np.ones(parts),
        "Kpart": Kpart,
        "Kwell": K_CG(rwell, KG, sig2, corr, e, prop, Kwell),
    }

    # call the stehfest-algorithm
    res = sf(grf_laplace, time, bound=stehfestn, **kwargs)

    # if the input are unstructured space-time points, return an array
    if not struc_grid and grid_shape:
        res = np.copy(np.diag(res).reshape(grid_shape))

    # add the reference head
    res += hinf

    return res


if __name__ == "__main__":
    import doctest

    doctest.testmod()
